;; BF Java Code Emitter
;; Jon Simons (simonsj at ccs dot neu dot edu)
;; August, 2007
;; -------------------------------------------------------------------------
;; This module emits Java code for the BF IR.
;;
;;   The emitted code is in the form of a "main" method in a Java class
;;   named "BF".  The method is declared to throw java.io.IOException,
;;   in case of out-of-bounds errors.
;;
;;   Memory is represented as a char[].  "The pointer" is represented via
;;   an index into memory and offsets.  Bounds checking is provided for
;;   free by the Java language.
;;
;;   System.out.print and System.in.read are used for IN and OUT emitting.
;;
;;   KNOWN LIMITATION:
;;
;;     The emitted code does not take into account the Java size
;;     restriction for class files.  A method's byte code is limited to
;;     64kb.  So if you emit a huge BF program and then try to compile with
;;     javac, you will receive a "code too large" error.  I discovered this
;;     nuance when trying to compile a huge BF program.
;;
;;     TODO:  This limitation will be fixed by simply mapping the IR into
;;            a list of "method chunks" (essentially basic blocks that can
;;            execute independently of one another, but execute correctly
;;            if they are executed in the correct order).
;;
;;   This module emits code which is completely oblivious to any integer
;;   wrapping issues.
;; -------------------------------------------------------------------------
(module bf-emit-java mzscheme
  
  (provide emit)
  
  ;; For AST structs
  (require "../bf-ast-to-ir.scm")
  ;; Emit Utils
  (require "bf-emit-util.scm")
  
  (require scheme/list)
  
  ;; emit : (list of EXPRs) -> string
  (define (emit AST)
    (string-append (emit-prologue)
                   (emit-program AST)
                   (emit-epilogue)))
  
  ;; emit-prologue : -> string
  (define (emit-prologue)
    (string-append
     "public class BF {\n"
     "  public static void main(String args[]) throws java.io.IOException {\n"
     "    char mem[] = new char[30000];\n"
     "    int i = 0;\n"))
  
  ;; emit-epilogue : -> string
  (define (emit-epilogue)
    (string-append "    return;~%"
                   "  }~%"
                   "}"))
  
  (define OUT-STRING "System.out.print(mem[i + ~v]);~%")
  (define IN-STRING  "mem[i + ~v] = (char) System.in.read();~%")
  
  ;; emit-program : (list of EXPRs) -> string
  (define (emit-program AST)
    (apply string-append (emit-exprs AST #f 4 ())))
  
  ;; Emits a list of strings which correspond to the given
  ;; list of EXPRs, using the given indentation level.
  ;;
  ;; The argument "mult-mode" should be #f for normal emitting,
  ;; or for MUL emitting, a MULTEMP.
  ;;
  ;; emit-exprs : (list of EXPRs) (#f or MULTEMP) number (list of strings)
  ;;              -> (list of strings)
  (define (emit-exprs lst mult-mode indent acc)
    ;; Here emit and emit+ are helpers for padding and formatting
    ;; strings to be emitted
    (let ((emit  (case-lambda
                   (()   (error "emitting nothing?"))
                   ((x)  (pad x indent))
                   (args (pad (apply format args) indent))))
          (emit+ (case-lambda
                   (()   (error "emitting nothing?"))
                   ((x)  (pad x (+ 2 indent)))
                   (args (pad (apply format args) (+ 2 indent))))))
      (if (null? lst)
          (reverse acc)
          (let* ((e   (car lst))
                 (ptr (EXPR-ptr e))
                 (N   (EXPR-N   e))
                 (str (case (EXPR-type e)
                        ((inc)  (emit "mem[i + ~v] += ~v~a;~%"
                                      ptr
                                      N
                                      (mult-string mult-mode)))
                        ((dec)  (emit "mem[i + ~v] -= ~v~a;~%"
                                      ptr
                                      N
                                      (mult-string mult-mode)))
                        ((lt)   (emit "i -= ~v;~%" N))
                        ((rt)   (emit "i += ~v;~%" N))
                        ((out)  (emit OUT-STRING ptr))
                        ((in)   (emit IN-STRING ptr))
                        ((loop) (string-append
                                 (emit "while (mem[i + ~v] != 0) {~%" ptr)
                                 (apply string-append
                                        (emit-exprs (EXPR-exprs e)
                                                    #f
                                                    (+ 2 indent)
                                                    ()))
                                 (emit "}~%")))
                        ((mul) (let ((mul-tmp (gen-MULTEMP ptr)))
                                 (string-append
                                  (emit "{~%")
                                  (emit+ (MULTEMP-init mul-tmp))
                                  (emit+ "if (~a != 0) {~%"
                                         (MULTEMP-val mul-tmp))
                                  (apply string-append
                                         (emit-exprs (EXPR-exprs e)
                                                     mul-tmp
                                                     (+ 4 indent)
                                                     ()))
                                  (emit+ "}~%")
                                  (emit "}~%"))))
                        ((zero) (emit "mem[i + ~v] = 0;~%" ptr))
                        (else  (error "unknown token, can't emit")))))
            (emit-exprs (cdr lst) mult-mode indent (cons str acc))))))
  
  ;; MULTEMP Stuff -----------------------------------------------------------
  
  ;; Used for declaring and referencing temp variables in the target language.
  ;;
  ;; make-MULTEMP : string string -> MULTEMP
  (define-struct MULTEMP (init val))
  
  (define (gen-wrap)
    (let ((tmp-num (box 0)))
      (lambda (ptr)
        (begin
          (set-box! tmp-num (+ 1 (unbox tmp-num)))
          (make-MULTEMP (format "int mulTmp_~v = mem[i + ~v];~%"
                                (unbox tmp-num)
                                ptr)
                        (format "mulTmp_~v" (unbox tmp-num)))))))
  
  ;; Generates a multiplication string for the given MULTEMP
  ;;
  ;; mult-string : MULTEMP -> string
  (define (mult-string multmp)
    (if multmp
        (string-append " * " (MULTEMP-val multmp))
        ""))
  
  ;; Generates a MULTEMP using the given pointer location.
  ;;
  ;; gen-MULTEMP : number -> MULTEMP
  (define gen-MULTEMP (gen-wrap))
  
  )
