;; BF MIPS Assembly Emitter
;; Jon Simons (simonsj at ccs dot neu dot edu)
;; August, 2007
;; -------------------------------------------------------------------------
;; This module emits MIPS Assembly for the BF IR.
;;
;;   A 30000 byte space is reserved in the .data section for use as
;;   the BF "memory".  A 2 byte scratch space is reserved for string
;;   reading and printing syscalls (which require null-terminated
;;   strings).
;;
;;   Registers are used as follows:
;;
;;     $t0 is the "pointer" to the BF "memory"
;;
;;     $t1 is used for scratch space for INCs and DECs and other stuff
;;
;;     $t2 is used for "mult temps", which are the factors by which INCs
;;         and DECs get multiplied inside of MUL bodies
;;
;;     $t3 is used to store the result of multiplications
;;
;;   Emitted code was tested using PCSpim on Windows, and SPIM on Linux.
;;
;;   This module emits code which is completely oblivious to any integer
;;   wrapping issues.
;; -------------------------------------------------------------------------
(module bf-emit-mips mzscheme
  
  (provide emit)
  
  ;; For AST structs
  (require "../bf-ast-to-ir.scm")
  ;; Emit Utils
  (require "bf-emit-util.scm")
  
  (require scheme/list)
  
  ;; emit : (list of EXPRs) -> string
  (define (emit AST)
    (string-append (emit-prologue)
                   (emit-program AST)
                   (emit-epilogue)))
  
  ;; emit-prologue : -> string
  (define (emit-prologue)
    (string-append "        .data~%"
                   "Scratch:~%"
                   "        .space 2~%"
                   "BFmemory:~%"
                   "        .space 30000~%"
                   "~%"
                   "        .text~%"
                   "main:~%"
                   "        la   $t0, BFmemory~%"
                   "~%"))
  
  ;; emit-epilogue : -> string
  (define (emit-epilogue)
    (string-append "        # syscall 10 (exit)~%"
                   "        li   $v0 10~%"
                   "        syscall"))
  
  ;; emit-program : (list of EXPRs) -> string
  (define (emit-program AST)
    (apply string-append (emit-exprs AST #f 8 ())))
  
  ;; Emits a list of strings which correspond to the given
  ;; list of EXPRs, using the given indentation level.
  ;;
  ;; The argument "mult-mode" should be #f for normal emitting,
  ;; and #t for multiplication emitting.
  ;;
  ;; emit-exprs : (list of EXPRs) boolean number (list of strings)
  ;;              -> (list of strings)
  (define (emit-exprs lst mult-mode indent acc)
    ;; Here emit is a helper for padding and formatting
    ;; strings to be emitted
    (let ((emit  (case-lambda
                   (()   (error "emitting nothing?"))
                   ((x)  (pad x indent))
                   (args (pad (apply format args) indent)))))
      (if (null? lst)
          (reverse acc)
          (let* ((e   (car lst))
                 (ptr (EXPR-ptr e))
                 (N   (EXPR-N   e))
                 (str (case (EXPR-type e)
                        ((inc)  (string-append
                                 (emit "lb   $t1, ~v($t0)~%" ptr)
                                 (if mult-mode
                                     (string-append
                                      (emit "mul  $t3, $t2, ~v~%" N)
                                      (emit "add  $t1, $t1, $t3~%"))
                                     (emit "addi $t1, ~v~%" N))
                                 (emit "sb   $t1, ~v($t0)~%" ptr)))
                        ((dec)  (string-append
                                 (emit "lb   $t1, ~v($t0)~%" ptr)
                                 (if mult-mode
                                     (string-append
                                      (emit "mul  $t3, $t2, ~v~%" N)
                                      (emit "sub  $t1, $t1, $t3~%"))
                                     (emit "addi $t1, -~v~%" N))
                                 (emit "sb   $t1, ~v($t0)~%" ptr)))
                        ((lt)   (emit "addi $t0, -~v~%" N))
                        ((rt)   (emit "addi $t0, ~v~%" N))
                        ((out)  (string-append
                                 (emit "lb   $a1, ~v($t0)~%" ptr)
                                 (emit "la   $a0, Scratch~%")
                                 (emit "sb   $a1, 0($a0)~%")
                                 (emit "li   $v0, 4~%")
                                 (emit "syscall # out~%")))
                        ((in)   (string-append
                                 (emit "la   $a0, Scratch~%")
                                 (emit "li   $a1, 2~%")
                                 (emit "li   $v0, 8~%")
                                 (emit "syscall # in~%")
                                 (emit "lb    $a0, 0($a0)~%")
                                 (emit "sb   $a0, ~v($t0)~%")))
                        ((loop)  (let* ((loop-tmp (gen-loop))
                                        (start    (GENLOOP-start loop-tmp))
                                        (end      (GENLOOP-end   loop-tmp)))
                                   (string-append
                                    (format "~a:~%" start)
                                    (emit "lb   $t1, ~v($t0)~%" ptr)
                                    (emit "beqz $t1, ~a~%" end)
                                    (apply string-append
                                           (emit-exprs (EXPR-exprs e)
                                                       #f
                                                       indent
                                                       ()))
                                    (emit "b    ~a~%" start)
                                    (format "~a:~%" end))))
                        ((mul)  (let ((skip-label (mul-skip)))
                                  (string-append
                                   (emit "lb   $t2, ~v($t0)~%" ptr)
                                   (emit "beqz $t2, ~a~%" skip-label)
                                   (apply string-append
                                          (emit-exprs (EXPR-exprs e)
                                                      #t
                                                      indent
                                                      ()))
                                   (format "~a:~%" skip-label))))
                        ((zero) (emit "sb   $zero, ~v($t0)~%" ptr))
                        (else  (error "unknown token, can't emit")))))
            (emit-exprs (cdr lst) mult-mode indent (cons str acc))))))
  
  )
